"""Definitions for dataclasses."""

# Do not lint for exception strings being assigned to variables first
# ruff: noqa: EM101
from datetime import datetime, timedelta
from typing import Self
from uuid import UUID

import pyarrow as pa
from dagster_pandera import pandera_schema_to_dagster_type
from pandera import DataFrameModel
from pandera.api.pandas.model_config import BaseConfig
from pandera.engines.pandas_engine import PydanticModel
from pydantic import BaseModel, Field

from horizon_data_core.helpers import name_to_uuid


class BasePyarrowModel(BaseModel):
    """A base model for PyArrow models."""

    @classmethod
    def pyarrow_schema(cls) -> pa.Schema:
        """Get the PyArrow schema for the model."""
        raise NotImplementedError("Subclasses must implement this method.")

    def to_pyarrow(self) -> pa.Table:
        """Convert the model to a PyArrow table."""
        return pa.Table.from_pylist([self.model_dump()], schema=self.pyarrow_schema())


class BasePostgresModel(BaseModel):
    """A base model for Postgres models.

    This simply ensures that all Postgres models have an id.
    """

    id: UUID | None
    created_datetime: None | datetime = None
    modified_datetime: None | datetime = None


class Mission(BasePostgresModel):
    """A group that can group entities around some bounded time or purpose."""

    name: None | str = None
    start_datetime: None | datetime = None
    end_datetime: None | datetime = None
    position: None | str = None
    free_text: None | str = None
    organization_id: UUID | None = None


class MissionEntity(BasePostgresModel):
    """A relationship table that specifies entity association with a mission."""

    mission_id: UUID
    entity_id: UUID
    organization_id: UUID | None = None


class EntityKind(BasePostgresModel):
    """A kind of entity."""

    name: None | str = None
    image_url: None | str = None
    short_description: None | str = None
    long_description: None | str = None
    organization_id: UUID | None = None


class Entity(BasePostgresModel):
    """An instance of an entity."""

    name: None | str = None
    kind_id: UUID
    free_text: None | str = None
    position: None | tuple[float, float] = None
    organization_id: UUID | None = None
    start_datetime: None | datetime = None
    end_datetime: None | datetime = None


class DataStream(BasePostgresModel):
    """A data stream of an entity."""

    entity_id: UUID
    organization_id: UUID | None = None
    name: None | str = None

    @classmethod
    def from_platform_and_name(cls, platform: Entity, data_stream_name: str) -> Self:
        """Create a DataStream from an entity and a name."""
        entity_id = platform.id
        assert entity_id is not None, "Entity ID should not be None when creating a datastream"
        name = data_stream_name.capitalize()
        id = name_to_uuid(f"{entity_id}:{name}")
        return cls(id=id, entity_id=entity_id, name=name)


class Ontology(BasePostgresModel):
    """An ontology for organizing entities."""

    name: None | str = None
    description: None | str = None
    organization_id: UUID | None = None


class OntologyClass(BasePostgresModel):
    """A class within an ontology."""

    ontology_id: UUID
    parent_id: UUID | None = None
    name: None | str = None
    description: None | str = None
    order: None | int = None
    organization_id: UUID | None = None
    relationship_type: None | str = None


class BeamgramSpecification(BasePostgresModel):
    """A specification for beamgram data processing."""

    center_bearing: None | float = None
    center_bin_width: None | float = None
    min_frequency: None | float = None
    max_frequency: None | float = None
    nfft: None | int = None
    organization_id: UUID | None = None
    elevation: None | float = None
    update_rate: None | timedelta = None
    normalizer: None | str = None


class BearingTimeRecordSpecification(BasePostgresModel):
    """A specification for bearing-time record data processing."""

    min_frequency: None | float = None
    max_frequency: None | float = None
    nfft: None | int = None
    frequency_spacing: None | float = None
    organization_id: UUID | None = None
    elevation: None | float = None
    update_rate: None | timedelta = None
    normalizer: None | str = None


class DataRow(BasePostgresModel, BasePyarrowModel):
    """A row of data in the Iceberg data_row table."""

    id: UUID | None = Field(default=None, exclude=True)
    data_stream_id: UUID
    datetime: datetime
    vector: list[float]
    data_type: str
    track_id: UUID
    vector_start_bound: float
    vector_end_bound: float

    @classmethod
    def pyarrow_schema(cls) -> pa.Schema:
        """Get the PyArrow schema for the model."""
        return pa.schema(
            (
                pa.field("data_stream_id", pa.string(), nullable=False),
                pa.field("datetime", pa.timestamp("us", tz="UTC"), nullable=False),
                pa.field("vector", pa.list_(pa.field("element", pa.float32(), nullable=False)), nullable=False),
                pa.field("data_type", pa.string(), nullable=False),
                pa.field("track_id", pa.string(), nullable=True),
                pa.field("vector_start_bound", pa.float32(), nullable=False),
                pa.field("vector_end_bound", pa.float32(), nullable=False),
            )
        )


class MetadataRow(BasePostgresModel, BasePyarrowModel):
    """A row of metadata in the Iceberg metadata_row table."""

    id: UUID | None = Field(default=None, exclude=True)
    data_stream_id: UUID
    datetime: datetime
    latitude: None | float = None
    longitude: None | float = None
    altitude: None | float = None
    speed: None | float = None
    heading: None | float = None
    pitch: None | float = None
    roll: None | float = None
    speed_over_ground: None | float = None

    @classmethod
    def pyarrow_schema(cls) -> pa.Schema:
        """Get the PyArrow schema for the model."""
        return pa.schema(
            (
                pa.field("data_stream_id", pa.string(), nullable=False),
                pa.field("datetime", pa.timestamp("us", tz="UTC"), nullable=False),
                pa.field("latitude", pa.float32(), nullable=True),
                pa.field("longitude", pa.float32(), nullable=True),
                pa.field("altitude", pa.float32(), nullable=True),
                pa.field("speed", pa.float32(), nullable=True),
                pa.field("heading", pa.float32(), nullable=True),
                pa.field("pitch", pa.float32(), nullable=True),
                pa.field("roll", pa.float32(), nullable=True),
                pa.field("speed_over_ground", pa.float32(), nullable=True),
            )
        )


class EntityBearingTimeRecordSpecification(BasePostgresModel):
    """A relationship table that specifies entity association with a btr specification."""

    entity_id: UUID
    bearing_time_record_specification_id: UUID
    organization_id: UUID | None = None


class EntityBeamgramSpecification(BasePostgresModel):
    """A relationship table that specifies entity association with a beamgram specification."""

    entity_id: UUID
    beamgram_specification_id: UUID
    organization_id: UUID | None = None


class EntityKindSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(EntityKind)
        title = "EntityKindSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class EntitySchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(Entity)
        title = "EntitySchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class DataStreamSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(DataStream)
        title = "DataStreamSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class MissionSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(Mission)
        title = "MissionSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class MissionEntitySchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(MissionEntity)
        title = "MissionEntitySchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class OntologySchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(Ontology)
        title = "OntologySchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class OntologyClassSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(OntologyClass)
        title = "OntologyClassSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class BeamgramSpecificationSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(BeamgramSpecification)
        title = "BeamgramSpecificationSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class BearingTimeRecordSpecificationSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(BearingTimeRecordSpecification)
        title = "BearingTimeRecordSpecificationSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class DataRowSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(DataRow)
        title = "DataRowSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


class MetadataRowSchema(DataFrameModel):
    """Pandera schema using the Pydantic model."""

    class Config(BaseConfig):
        """Config with dataframe-level data type."""

        dtype = PydanticModel(MetadataRow)
        title = "MetadataRowSchema"
        coerce = True  # This is required, otherwise a SchemaInitError is raised


EntityKindSchemaType = pandera_schema_to_dagster_type(EntityKindSchema)
EntitySchemaType = pandera_schema_to_dagster_type(EntitySchema)
DataStreamSchemaType = pandera_schema_to_dagster_type(DataStreamSchema)
MissionSchemaType = pandera_schema_to_dagster_type(MissionSchema)
MissionEntitySchemaType = pandera_schema_to_dagster_type(MissionEntitySchema)
OntologySchemaType = pandera_schema_to_dagster_type(OntologySchema)
OntologyClassSchemaType = pandera_schema_to_dagster_type(OntologyClassSchema)
BeamgramSpecificationSchemaType = pandera_schema_to_dagster_type(BeamgramSpecificationSchema)
BearingTimeRecordSpecificationSchemaType = pandera_schema_to_dagster_type(BearingTimeRecordSpecificationSchema)
DataRowSchemaType = pandera_schema_to_dagster_type(DataRowSchema)
MetadataRowSchemaType = pandera_schema_to_dagster_type(MetadataRowSchema)



"""IO Managers for Dagster.

This moddle includes:
- PostgresDataFrameIOManager: for writing Pandas DFs to Postgres tables and back.
"""

import math
from collections.abc import Mapping
from typing import Any, Final, Self, TypeGuard

from dagster import ConfigurableIOManager, InputContext, OutputContext, get_dagster_logger
from pandas import DataFrame, read_sql
from sqlalchemy import MetaData, Table, create_engine
from sqlalchemy.engine import Engine

from .database import construct_query_statement, create_upsert_method

logger = get_dagster_logger()

# Number of coordinates in a PostgreSQL point (longitude, latitude)
POINT_COORDINATE_COUNT: Final[int] = 2


class PostgresDataframeIOManager(ConfigurableIOManager):
    """An IO manager to read/write to or from Pandas dataframes and Postgres tables."""

    username: str = "postgres"
    password: str = "postgres"  # noqa: S105
    hostname: str = "localhost"
    port: int = 5432
    database_name: str = "postgres"
    ssl_mode: str | None = None
    if_exists: str = "append"
    query_columns_list: list[str] | None = None
    query_predicate: str | None = None

    def _load_input_context_metadata(
        self: Self,
        context: InputContext,
    ) -> Mapping[str, Any]:
        """Load and validate input context metadata."""
        upstream_metadata: Mapping[str, Any] = {}
        input_metadata: Mapping[str, Any] = {}

        if context.upstream_output and context.upstream_output.metadata:
            upstream_metadata = context.upstream_output.metadata

        if context.metadata:
            input_metadata = context.metadata

        metadata = {**upstream_metadata, **input_metadata}

        if not metadata:
            msg = "upstream output or input metadata must be defined."
            raise ValueError(msg)

        return metadata

    def _load_output_context_metadata(
        self: Self,
        context: OutputContext,
    ) -> Mapping[str, Any]:
        """Load and validate output context metadata."""
        if not context.metadata:
            msg = "context.metadata cannot be `None`."
            raise ValueError(msg)
        return context.metadata

    def _get_schema(self: Self, metadata: Mapping[str, Any]) -> str:
        """Get a PostgreSQL schema from a Dagster context."""
        schema_name = metadata.get("schema_name")
        if schema_name is None:
            msg = "'schema_name' cannot be `None`."
            raise ValueError(msg)
        if not isinstance(schema_name, str):
            msg = f"Expected 'schema_name' to be of type str. Got {type(schema_name)} instead."
            raise TypeError(msg)
        return schema_name

    def _get_table_name(self: Self, metadata: Mapping[str, Any]) -> str:
        """Get a PostgreSQL table name from a Dagster context."""
        table_name = metadata.get("table_name")
        if table_name is None:
            msg = "'table_name' cannot be `None`."
            raise ValueError(msg)
        if not isinstance(table_name, str):
            msg = f"Expected 'table_name' to be of type str. Got {type(table_name)} instead."
            raise TypeError(msg)
        return table_name

    def _get_table_path(
        self: "PostgresDataframeIOManager",
        metadata: Mapping[str, Any],
    ) -> tuple[str, str]:
        """Get a PostgreSQL schema and table name from a Dagster context."""
        schema_name = self._get_schema(metadata)
        table_name = self._get_table_name(metadata)
        return (schema_name, table_name)

    def _get_connection_string(
        self: "PostgresDataframeIOManager",
        metadata: Mapping[str, Any],
    ) -> str:
        """Return the postgres connection string from context metadata."""
        username = metadata.get("username", self.username)
        hostname = metadata.get("hostname", self.hostname)
        database_name = metadata.get("database_name", self.database_name)
        port = metadata.get("port", self.port)
        return f"postgresql://{username}:{self.password}@{hostname}:{port}/{database_name}"

    def _get_engine(self: "PostgresDataframeIOManager", metadata: Mapping[str, Any]) -> Engine:
        ssl_mode = metadata.get("ssl_mode", self.ssl_mode)
        connect_args = {}
        # Psycopg2 does not support sslmode=no-verify, so we need to use a dummy sslrootcert for dev mode connections
        if ssl_mode == "no-verify":
            connect_args = {"sslmode": "require", "sslrootcert": "/path/to/empty/file"}
        else:
            connect_args = {"sslmode": ssl_mode}
        return create_engine(self._get_connection_string(metadata), connect_args=connect_args)

    def _get_table(self: "PostgresDataframeIOManager", metadata: Mapping[str, Any], engine: Engine) -> Table:
        (schema_name, table_name) = self._get_table_path(metadata)
        metadata_obj = MetaData(schema=schema_name)
        return Table(table_name, metadata_obj, autoload_with=engine)

    def _get_join(
        self: "PostgresDataframeIOManager",
        metadata: Mapping[str, Any],
        engine: Engine,
    ) -> list[tuple[Table, str]] | None:
        join_query = metadata.get("join", None)
        if not join_query:
            return None

        if is_list_of_str_tuples(join_query) is False:
            msg = (
                f"Expected 'join_query' to be of type list[tuple[str,str]]."
                f"Got {join_query}, type:{type(join_query)} instead."
            )
            raise TypeError(msg)

        schema_name = self._get_schema(metadata)
        metadata_obj = MetaData(schema=schema_name)
        join_list: list[tuple[Table, str]] = []
        for join_name, on in join_query:
            join_table = Table(join_name, metadata_obj, autoload_with=engine)
            join_list.append((join_table, on))
        return join_list

    def handle_output(
        self: "PostgresDataframeIOManager",
        context: OutputContext,
        dataframe: DataFrame | None,
    ) -> None:
        """Return the number of DB rows modified or `None`.

        Write the supplied dataframe object to postgres and create the target table from a schema definition if
        the table does not already exist.
        """
        if dataframe is None:
            return

        metadata = self._load_output_context_metadata(context)
        schema_name, table_name = self._get_table_path(metadata)
        schema_table_name = f"{schema_name}.{table_name}"
        engine = self._get_engine(metadata)
        should_upsert = metadata.get("should_upsert")
        for column in dataframe.columns:
            if dataframe[column].dtype == "timedelta64[ns]":  # pragma: no cover
                dataframe[column] = dataframe[column].apply(str)
            # Convert tuple positions to PostgreSQL point string format: "(lon,lat)"
            elif column == "position" and dataframe[column].dtype == "object" and len(dataframe) > 0:
                # Check if values are tuples (for point type)
                sample_value = dataframe[column].iloc[0]
                if isinstance(sample_value, tuple) and len(sample_value) == POINT_COORDINATE_COUNT:

                    def convert_to_point(x: tuple[float, float] | None | float) -> str | None:
                        if isinstance(x, tuple) and len(x) == POINT_COORDINATE_COUNT:
                            return f"({x[0]},{x[1]})"
                        if x is None or (isinstance(x, float) and math.isnan(x)):
                            return None
                        msg = f"Expected tuple of length {POINT_COORDINATE_COUNT} or None, got {type(x)}: {x}"
                        raise TypeError(msg)

                    dataframe[column] = dataframe[column].apply(convert_to_point)
        rows_affected = dataframe.to_sql(
            name=table_name,
            schema=schema_name,
            con=engine,
            index=False,
            if_exists="append",
            method=create_upsert_method(schema_name, table_name) if should_upsert else None,
        )

        context.add_output_metadata(
            {
                "db": self.database_name,
                "schema_table_name": schema_table_name,
                "dataframe_rows": len(dataframe),
                "records_affected": rows_affected,
            },
        )

    def load_input(
        self: "PostgresDataframeIOManager",
        context: InputContext,
    ) -> DataFrame:
        """Run a postgres query and converts the result back into a DataFrame."""
        metadata = self._load_input_context_metadata(context)
        engine = self._get_engine(metadata)
        table = self._get_table(metadata, engine)
        query_columns_list = metadata.get("columns")
        query_join = self._get_join(metadata, engine)
        query_predicate = metadata.get("predicate", lambda _: None)(context)
        statement = construct_query_statement(table, query_columns_list, query_join, query_predicate)
        logger.info("Executing: %s", statement)
        return read_sql(statement, engine)


def is_list_of_str_tuples(value: Any) -> TypeGuard[list[tuple[str, str]]]:  # noqa: ANN401
    """Typecheck of the value is a list of tuples of 2 strings."""
    return isinstance(value, list) and all(
        isinstance(t, tuple) and len(t) == 2 and all(isinstance(e, str) for e in t)  # noqa: PLR2004
        for t in value
    )



"""Client for Postgres operations."""

from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from functools import cached_property
from typing import Protocol

from sqlalchemy import URL, Engine, create_engine
from sqlalchemy.orm import Session, sessionmaker


class Client(Protocol):
    """Client protocol for database operations."""

    @property
    def connection_string(self) -> URL:
        """The connection string for the Postgres client."""
        ...

    @cached_property
    def engine(self) -> Engine:
        """The engine for the Postgres client."""
        ...

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """Get a session for the Postgres client."""
        ...


@dataclass
class PostgresClient:
    """Client for Postgres operations."""

    user: str
    password: str
    host: str
    port: str | int
    database: str
    dialect: str = "postgresql"
    driver: str = "psycopg2"
    sslmode: str = "require"
    channel_binding: str = "require"

    @property
    def connection_string(self) -> URL:
        """Get the connection string for the Postgres client."""
        return URL.create(
            drivername=f"{self.dialect}+{self.driver}",
            username=self.user,
            password=self.password,
            host=self.host,
            port=int(self.port),
            database=self.database,
            query={
                "sslmode": self.sslmode,
                "channel_binding": self.channel_binding,
            },
        )

    @cached_property
    def engine(self) -> Engine:
        """The engine for the Postgres client."""
        return create_engine(
            self.connection_string,
            connect_args={
                "sslmode": self.sslmode,
                "channel_binding": self.channel_binding,
            },
        )

    @cached_property
    def _session_factory(self) -> sessionmaker:
        return sessionmaker(bind=self.engine)

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """Get a session for the Postgres client."""
        session = self._session_factory()
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
